"""Run all test cases."""

import sys

import numpy as np
from packaging.version import Version

import tables as tb
from tables.tests import common
from tables.tests.test_suite import suite, test


def get_tuple_version(hexversion):
    """Get a tuple from a compact version in hex."""

    h = hexversion
    return(h & 0xff0000) >> 16, (h & 0xff00) >> 8, h & 0xff


if __name__ == '__main__':

    common.parse_argv(sys.argv)

    hdf5_version = get_tuple_version(tb.which_lib_version("hdf5")[0])
    hdf5_version_str = "%s.%s.%s" % hdf5_version
    if Version(hdf5_version_str) < tb.req_versions.min_hdf5_version:
        print(f"*Warning*: HDF5 version is lower than recommended: "
              f"{hdf5_version} < {tb.req_versions.min_hdf5_version}")

    if Version(np.__version__) < tb.req_versions.min_numpy_version:
        print(f"*Warning*: NumPy version is lower than recommended: "
              f"{np.__version__} < {tb.req_versions.min_numpy_version}")

    # Handle some global flags (i.e. only useful for test_all.py)
    only_versions = 0
    args = sys.argv[:]
    for arg in args:
        # Remove 'show-versions' for PyTables 2.3 or higher
        if arg in ['--print-versions', '--show-versions']:
            only_versions = True
            sys.argv.remove(arg)
        elif arg == '--show-memory':
            common.show_memory = True
            sys.argv.remove(arg)

    common.print_versions()
    if not only_versions:
        common.print_heavy(common.heavy)
        common.unittest.main(defaultTest='tb.tests.suite')



"""Here is defined the IndexArray class."""

from bisect import bisect_left, bisect_right

from .node import NotLoggedMixin
from .carray import CArray
from .earray import EArray
from . import indexesextension


# Declarations for inheriting


class CacheArray(indexesextension.CacheArray, NotLoggedMixin, EArray):
    """Container for keeping index caches of 1st and 2nd level."""

    # Class identifier.
    _c_classid = 'CACHEARRAY'


class LastRowArray(indexesextension.LastRowArray, NotLoggedMixin, CArray):
    """Container for keeping sorted and indices values of last row of an
    index."""

    # Class identifier.
    _c_classid = 'LASTROWARRAY'


class IndexArray(indexesextension.IndexArray, NotLoggedMixin, EArray):
    """Represent the index (sorted or reverse index) dataset in HDF5 file.

    All NumPy typecodes are supported except for complex datatypes.

    Parameters
    ----------
    parentnode
        The Index class from which this object will hang off.

        .. versionchanged:: 3.0
           Renamed from *parentNode* to *parentnode*.

    name : str
        The name of this node in its parent group.
    atom
        An Atom object representing the shape and type of the atomic objects to
        be saved. Only scalar atoms are supported.
    title
        Sets a TITLE attribute on the array entity.
    filters : Filters
        An instance of the Filters class that provides information about the
        desired I/O filters to be applied during the life of this object.
    byteorder
        The byteroder of the data on-disk.

    """

    # Class identifier.
    _c_classid = 'INDEXARRAY'

    @property
    def chunksize(self):
        """The chunksize for this object."""
        return self.chunkshape[1]

    @property
    def slicesize(self):
        """The slicesize for this object."""
        return self.shape[1]

    def __init__(self, parentnode, name,
                 atom=None, title="",
                 filters=None, byteorder=None):
        """Create an IndexArray instance."""

        self._v_pathname = parentnode._g_join(name)
        if atom is not None:
            # The shape and chunkshape needs to be fixed here
            if name == "sorted":
                reduction = parentnode.reduction
                shape = (0, parentnode.slicesize // reduction)
                chunkshape = (1, parentnode.chunksize // reduction)
            else:
                shape = (0, parentnode.slicesize)
                chunkshape = (1, parentnode.chunksize)
        else:
            # The shape and chunkshape will be read from disk later on
            shape = None
            chunkshape = None

        super().__init__(
            parentnode, name, atom, shape, title, filters,
            chunkshape=chunkshape, byteorder=byteorder)

    # This version of searchBin uses both ranges (1st level) and
    # bounds (2nd level) caches. It uses a cache for boundary rows,
    # but not for 'sorted' rows (this is only supported for the
    # 'optimized' types).
    def _search_bin(self, nrow, item):
        item1, item2 = item
        result1 = -1
        result2 = -1
        hi = self.shape[1]
        ranges = self._v_parent.rvcache
        boundscache = self.boundscache
        # First, look at the beginning of the slice
        begin = ranges[nrow, 0]
        # Look for items at the beginning of sorted slices
        if item1 <= begin:
            result1 = 0
        if item2 < begin:
            result2 = 0
        if result1 >= 0 and result2 >= 0:
            return (result1, result2)
        # Then, look for items at the end of the sorted slice
        end = ranges[nrow, 1]
        if result1 < 0:
            if item1 > end:
                result1 = hi
        if result2 < 0:
            if item2 >= end:
                result2 = hi
        if result1 >= 0 and result2 >= 0:
            return (result1, result2)
        # Finally, do a lookup for item1 and item2 if they were not found
        # Lookup in the middle of slice for item1
        chunksize = self.chunksize  # Number of elements/chunksize
        nchunk = -1
        # Try to get the bounds row from the LRU cache
        nslot = boundscache.getslot(nrow)
        if nslot >= 0:
            # Cache hit. Use the row kept there.
            bounds = boundscache.getitem(nslot)
        else:
            # No luck with cached data. Read the row and put it in the cache.
            bounds = self._v_parent.bounds[nrow]
            size = bounds.size * bounds.itemsize
            boundscache.setitem(nrow, bounds, size)
        if result1 < 0:
            # Search the appropriate chunk in bounds cache
            nchunk = bisect_left(bounds, item1)
            chunk = self._read_sorted_slice(nrow, chunksize * nchunk,
                                            chunksize * (nchunk + 1))
            result1 = indexesextension._bisect_left(chunk, item1, chunksize)
            result1 += chunksize * nchunk
        # Lookup in the middle of slice for item2
        if result2 < 0:
            # Search the appropriate chunk in bounds cache
            nchunk2 = bisect_right(bounds, item2)
            if nchunk2 != nchunk:
                chunk = self._read_sorted_slice(nrow, chunksize * nchunk2,
                                                chunksize * (nchunk2 + 1))
            result2 = indexesextension._bisect_right(chunk, item2, chunksize)
            result2 += chunksize * nchunk2
        return (result1, result2)

    def __str__(self):
        """A compact representation of this class"""
        return f"IndexArray(path={self._v_pathname})"

    def __repr__(self):
        """A verbose representation of this class."""

        return f"""{self}
  atom = {self.atom!r}
  shape = {self.shape}
  nrows = {self.nrows}
  chunksize = {self.chunksize}
  slicesize = {self.slicesize}
  byteorder = {self.byteorder!r}"""



########################################################################
#
# License: BSD
# Created: June 17, 2005
# Author:  Francesc Alted - faltet@pytables.com
#
# $Id$
#
########################################################################

"""Here is where Table and Row extension types live.

Classes (type extensions):

    Table
    Row

Functions:

Misc variables:

"""

import sys
import numpy
from time import time
import platform

from .description import Col
from .exceptions import HDF5ExtError
from .conditions import call_on_recarr
from .utilsextension import (get_nested_field, atom_from_hdf5_type,
  create_nested_type, hdf5_to_np_ext_type, create_nested_type, platform_byteorder,
  pttype_to_hdf5, pt_special_kinds, npext_prefixes_to_ptkinds, hdf5_class_to_string,
  H5T_STD_I64)
from .utils import SizeType

from .utilsextension cimport get_native_type, cstr_to_pystr

# numpy functions & objects
from hdf5extension cimport Leaf
from cpython cimport PyErr_Clear
from libc.stdio cimport snprintf
from libc.stdlib cimport malloc, free
from libc.string cimport memcpy, strdup, strcmp, strlen
from numpy cimport (import_array, ndarray, npy_intp, PyArray_GETITEM,
  PyArray_SETITEM, PyArray_BYTES, PyArray_DATA, PyArray_NDIM, PyArray_STRIDE)
from .definitions cimport (hid_t, herr_t, hsize_t, htri_t, hbool_t,
  H5F_ACC_RDONLY, H5P_DEFAULT, H5D_CHUNKED, H5T_DIR_DEFAULT,
  H5F_SCOPE_LOCAL, H5F_SCOPE_GLOBAL, H5T_COMPOUND, H5Tget_order,
  H5Fflush, H5Dget_create_plist, H5T_ORDER_LE,
  H5D_layout_t, H5Dopen, H5Dclose, H5Dread, H5Dget_type, H5Dget_space,
  H5Pget_layout, H5Pget_chunk, H5Pclose,
  H5Sget_simple_extent_ndims, H5Sget_simple_extent_dims, H5Sclose,
  H5T_class_t, H5Tget_size, H5Tset_size, H5Tcreate, H5Tcopy, H5Tclose,
  H5Tget_nmembers, H5Tget_member_name, H5Tget_member_type, H5Tget_native_type,
  H5Tget_member_offset, H5Tinsert, H5Tget_class, H5Tget_super, H5Tget_offset,
  H5T_cset_t, H5T_CSET_ASCII, H5T_CSET_UTF8,
  H5ATTRset_attribute_string, H5ATTRset_attribute,
  get_len_of_range, get_order, set_order, is_complex,
  conv_float64_timeval32, truncate_dset,
  pt_H5free_memory)

from .lrucacheextension cimport ObjectCache, NumCache



#-----------------------------------------------------------------

# Optimized HDF5 API for PyTables
cdef extern from "H5TB-opt.h" nogil:

  herr_t H5TBOmake_table( char *table_title, hid_t loc_id, char *dset_name,
                          char *version, char *class_,
                          hid_t mem_type_id, hsize_t nrecords,
                          hsize_t chunk_size, hsize_t block_size,
                          void *fill_data, int compress,
                          char *complib, int shuffle, int fletcher32,
                          hbool_t track_times, hbool_t blosc2_support,
                          void *data )

  herr_t H5TBOread_records( char* filename, hbool_t blosc2_support,
                            hid_t dataset_id, hid_t mem_type_id,
                            hsize_t start, hsize_t nrecords, void *data )

  herr_t H5TBOread_elements( hid_t dataset_id, hid_t mem_type_id,
                             hsize_t nrecords, void *coords, void *data )

  herr_t H5TBOappend_records( hbool_t blosc2_support, hid_t dataset_id,
                              hid_t mem_type_id, hsize_t start,
                              hsize_t nrecords, void *data )

  herr_t H5TBOwrite_records ( hbool_t blosc2_support, hid_t dataset_id,
                              hid_t mem_type_id, hsize_t start,
                              hsize_t nrecords, hsize_t step, void *data )

  herr_t write_records_blosc2( hid_t dataset_id, hid_t mem_type_id,
                               hsize_t start, hsize_t nrecords,
                               const void *data )

  herr_t H5TBOwrite_elements( hid_t dataset_id, hid_t mem_type_id,
                              hsize_t nrecords, void *coords, void *data )

  herr_t H5TBOdelete_records( char* filename, hbool_t blosc2_support,
                              hid_t   dataset_id, hid_t   mem_type_id,
                              hsize_t ntotal_records, size_t  src_size,
                              hsize_t start, hsize_t nrecords,
                              hsize_t maxtuples )


#----------------------------------------------------------------------------

# Initialization code

# The numpy API requires this function to be called before
# using any numpy facilities in an extension module.
import_array()

#-------------------------------------------------------------


# Private functions
cdef get_nested_field_cache(recarray, fieldname, fieldcache):
  """Get the maybe nested field named `fieldname` from the `recarray`.

  The `fieldname` may be a simple field name or a nested field name with
  slah-separated components. It can also be an integer specifying the position
  of the field.

  """

  try:
    field = fieldcache[fieldname]
  except KeyError:
    # Check whether fieldname is an integer and if so, get the field
    # straight from the recarray dictionary (it can't be anywhere else)
    if isinstance(fieldname, int):
      field = recarray[fieldname]
    else:
      field = get_nested_field(recarray, fieldname)
    fieldcache[fieldname] = field
  return field


cdef join_path(object parent, object name):
  if parent == "":
    return name
  else:
    return parent + '/' + name


# Public classes

cdef class Table(Leaf):
  # instance variables
  cdef void     *wbuf

  def _create_table(self, title, complib, obversion):
    cdef int     offset
    cdef int     ret
    cdef long    buflen
    cdef hid_t   oid
    cdef void    *data
    cdef hsize_t nrows
    cdef bytes   class_
    cdef ndarray wdflts
    cdef void    *fill_data
    cdef ndarray recarr
    cdef object  name
    cdef bytes encoded_title, encoded_complib, encoded_obversion
    cdef char *ctitle = NULL
    cdef char *cobversion = NULL
    cdef bytes encoded_name
    cdef char fieldname[128]
    cdef int i
    cdef H5T_cset_t cset = H5T_CSET_ASCII

    encoded_title = title.encode('utf-8')
    encoded_complib = complib.encode('utf-8')
    encoded_obversion = obversion.encode('utf-8')
    encoded_name = self.name.encode('utf-8')

    # Get the C pointer
    ctitle = encoded_title
    cobversion = encoded_obversion

    # Compute the complete compound datatype based on the table description
    self.disk_type_id = create_nested_type(self.description, self.byteorder)
    #self.type_id = H5Tcopy(self.disk_type_id)
    # A H5Tcopy only is not enough, as we want the in-memory type to be
    # in the byteorder of the machine (sys.byteorder).
    self.type_id = create_nested_type(self.description, sys.byteorder)

    # The fill values area
    wdflts = self._v_wdflts
    if wdflts is None:
      fill_data = NULL
    else:
      fill_data = PyArray_DATA(wdflts)

    # test if there is data to be saved initially
    if self._v_recarray is not None:
      recarr = self._v_recarray
      data = PyArray_DATA(recarr)
    else:
      data = NULL

    class_ = self._c_classid.encode('utf-8')
    cdef hsize_t blocksize = self._v_blocksize if hasattr(self, "_v_blocksize") else 0
    cdef hbool_t blosc2_support = ((self.byteorder == sys.byteorder) and
                                   (self.filters.complib != None) and
                                   (self.filters.complib[0:6] == "blosc2"))
    self.dataset_id = H5TBOmake_table(ctitle, self.parent_id, encoded_name,
                                      cobversion, class_, self.disk_type_id,
                                      self.nrows, self.chunkshape[0],
                                      blocksize, fill_data,
                                      self.filters.complevel, encoded_complib,
                                      self.filters.shuffle_bitshuffle,
                                      self.filters.fletcher32,
                                      self._want_track_times,
                                      blosc2_support, data)
    if self.dataset_id < 0:
      raise HDF5ExtError("Problems creating the table")

    if self._v_file.params['PYTABLES_SYS_ATTRS']:
      cset = H5T_CSET_UTF8
      # Set the conforming table attributes
      # Attach the CLASS attribute
      ret = H5ATTRset_attribute_string(self.dataset_id, "CLASS", class_,
                                       len(class_), cset)
      if ret < 0:
        raise HDF5ExtError("Can't set attribute '%s' in table:\n %s." %
                           ("CLASS", self.name))
      # Attach the VERSION attribute
      ret = H5ATTRset_attribute_string(self.dataset_id, "VERSION", cobversion,
                                       len(encoded_obversion), cset)
      if ret < 0:
        raise HDF5ExtError("Can't set attribute '%s' in table:\n %s." %
                           ("VERSION", self.name))
      # Attach the TITLE attribute
      ret = H5ATTRset_attribute_string(self.dataset_id, "TITLE", ctitle,
                                       len(encoded_title), cset)
      if ret < 0:
        raise HDF5ExtError("Can't set attribute '%s' in table:\n %s." %
                           ("TITLE", self.name))
      # Attach the NROWS attribute
      nrows = self.nrows
      ret = H5ATTRset_attribute(self.dataset_id, "NROWS", H5T_STD_I64,
                                0, NULL, <char *>&nrows)
      if ret < 0:
        raise HDF5ExtError("Can't set attribute '%s' in table:\n %s." %
                           ("NROWS", self.name))

      # Attach the FIELD_N_NAME attributes
      # We write only the first level names
      for i, name in enumerate(self.description._v_names):
        snprintf(fieldname, 128, "FIELD_%d_NAME", i)
        encoded_name = name.encode('utf-8')
        ret = H5ATTRset_attribute_string(self.dataset_id, fieldname,
                                         encoded_name, len(encoded_name),
                                         cset)
      if ret < 0:
        raise HDF5ExtError("Can't set attribute '%s' in table:\n %s." %
                           (fieldname, self.name))

    # If created in PyTables, the table is always chunked
    self._chunked = True  # Accessible from python

    # Finally, return the object identifier.
    return self.dataset_id


  cdef get_nested_type(self, hid_t type_id, hid_t native_type_id,
                       object colpath, object field_byteorders):
    """Open a nested type and return a nested dictionary as description."""

    cdef hid_t   member_type_id, native_member_type_id, member_offset
    cdef hsize_t nfields, i
    cdef hsize_t dims[1]
    cdef size_t  itemsize
    cdef char    *c_colname
    cdef H5T_class_t class_id
    cdef char    c_byteorder2[11]  # "irrelevant" fits easily here
    cdef char    *sys_byteorder
    cdef object  desc, colobj, colpath2, typeclassname, typeclass
    cdef object  byteorder
    cdef str     colname, byteorder2

    offset = 0
    desc = {}
    # Get the number of members
    nfields = H5Tget_nmembers(type_id)

    # Iterate through fields to get the correct order that elements may appear in
    # The object type can be stored not in order, so order based on the offset in the data
    position_order = []
    for i in range(nfields):
      member_offset = H5Tget_member_offset(type_id, i)
      position_order.append((member_offset, i))

    position_order.sort()

    # Iterate thru the members
    for pos, i in enumerate([x[1] for x in position_order]):
      # Get the member name
      c_colname = H5Tget_member_name(type_id, i)
      colname = cstr_to_pystr(c_colname)

      # Get the member type
      member_type_id = H5Tget_member_type(type_id, i)
      # Get the member offset
      member_offset = H5Tget_member_offset(type_id, i)
      # Get the HDF5 class
      class_id = H5Tget_class(member_type_id)
      if class_id == H5T_COMPOUND and not is_complex(member_type_id):
        colpath2 = join_path(colpath, colname)
        # Create the native data in-memory
        itemsize = H5Tget_size(member_type_id)
        native_member_type_id = H5Tcreate(H5T_COMPOUND, itemsize)
        desc[colname], itemsize = self.get_nested_type(
          member_type_id, native_member_type_id, colpath2, field_byteorders)
        desc[colname]["_v_pos"] = pos
        desc[colname]["_v_offset"] = member_offset
      else:
        # Get the member format and the corresponding Col object
        try:
          native_member_type_id = get_native_type(member_type_id)
          atom = atom_from_hdf5_type(native_member_type_id)
          colobj = Col.from_atom(atom, pos=pos, _offset=member_offset)
          itemsize = H5Tget_size(native_member_type_id)
        except TypeError, te:
          # Re-raise TypeError again with more info
          raise TypeError(
            ("table ``%s``, column ``%s``: %%s" % (self.name, colname))
            % te.args[0])
        desc[colname] = colobj
        # For time kinds, save the byteorder of the column
        # (useful for conversion of time datatypes later on)
        if colobj.kind == "time":
          colobj._byteorder = H5Tget_order(member_type_id)
          if colobj._byteorder == H5T_ORDER_LE:
            field_byteorders.append("little")
          else:
            field_byteorders.append("big")
        elif colobj.kind in ['int', 'uint', 'float', 'complex', 'enum']:
          # Keep track of the byteorder for this column
          get_order(member_type_id, c_byteorder2)
          byteorder2 = cstr_to_pystr(c_byteorder2)
          if byteorder2 in ["little", "big"]:
            field_byteorders.append(byteorder2)

      # Insert the native member
      H5Tinsert(native_type_id, c_colname, member_offset, native_member_type_id)
      # Update the offset
      offset = offset + itemsize
      # Release resources
      H5Tclose(native_member_type_id)
      H5Tclose(member_type_id)
      pt_H5free_memory(c_colname)

    # set the byteorder and other things (just in top level)
    if colpath == "":
      # Compute a byteorder for the entire table
      if len(field_byteorders) > 0:
        field_byteorders = numpy.array(field_byteorders)
        # Cython doesn't interpret well the extended comparison
        # operators so this: field_byteorders == "little" doesn't work
        # as expected
        if numpy.alltrue(field_byteorders.__eq__("little")):
          byteorder = "little"
        elif numpy.alltrue(field_byteorders.__eq__("big")):
          byteorder = "big"
        else:  # Yes! someone has done it!
          byteorder = "mixed"
      else:
        byteorder = "irrelevant"
      self.byteorder = byteorder

    return desc, offset

  def _get_info(self):
    """Get info from a table on disk."""

    cdef hid_t   space_id, plist
    cdef size_t  type_size, size2
    cdef hsize_t dims[1]        # enough for unidimensional tables
    cdef hsize_t chunksize[1]
    cdef H5D_layout_t layout
    cdef bytes encoded_name

    encoded_name = self.name.encode('utf-8')

    # Open the dataset
    self.dataset_id = H5Dopen(self.parent_id, encoded_name, H5P_DEFAULT)
    if self.dataset_id < 0:
      raise HDF5ExtError("Non-existing node ``%s`` under ``%s``" %
                         (self.name, self._v_parent._v_pathname))

    # Get the datatype on disk
    self.disk_type_id = H5Dget_type(self.dataset_id)
    if H5Tget_class(self.disk_type_id) != H5T_COMPOUND:
        raise ValueError("Node ``%s`` is not a Table object" %
                         (self._v_parent._v_leaves[self.name]._v_pathname))
    # Get the number of rows
    space_id = H5Dget_space(self.dataset_id)
    H5Sget_simple_extent_dims(space_id, dims, NULL)
    self.nrows = SizeType(dims[0])
    # Free resources
    H5Sclose(space_id)

    # Get the layout of the datatype
    plist = H5Dget_create_plist(self.dataset_id)
    layout = H5Pget_layout(plist)
    if layout == H5D_CHUNKED:
      self._chunked = 1
      # Get the chunksize
      H5Pget_chunk(plist, 1, chunksize)
    else:
      self._chunked = 0
      chunksize[0] = 0
    H5Pclose(plist)

    # Get the type size
    type_size = H5Tget_size(self.disk_type_id)
    # Create the native data in-memory
    self.type_id = H5Tcreate(H5T_COMPOUND, type_size)
    # Fill-up the (nested) native type and description
    desc, offset = self.get_nested_type(self.disk_type_id, self.type_id, "", [])

    if desc == {}:
      raise HDF5ExtError("Problems getting desciption for table %s", self.name)

    if offset < type_size:
      # Trailing padding, set the itemsize to the correct type_size (see #765)
      desc['_v_itemsize'] = type_size


    # Return the object ID and the description
    return (self.dataset_id, desc, SizeType(chunksize[0]))

  cdef _convert_time64_(self, ndarray nparr, hsize_t nrecords, int sense):
    """Converts a NumPy of Time64 elements between NumPy and HDF5 formats.

    NumPy to HDF5 conversion is performed when 'sense' is 0.  Otherwise, HDF5
    to NumPy conversion is performed.  The conversion is done in place,
    i.e. 'nparr' is modified.

    """

    cdef void *t64buf
    cdef long byteoffset
    cdef npy_intp bytestride, nelements

    byteoffset = 0   # NumPy objects doesn't have an offset
    bytestride = PyArray_STRIDE(nparr, 0)  # supports multi-dimensional recarray
    # Compute the number of elements in the multidimensional cell
    nelements = nparr.size // len(nparr)
    t64buf = PyArray_DATA(nparr)

    conv_float64_timeval32(
      t64buf, byteoffset, bytestride, nrecords, nelements, sense)

  cpdef _convert_types(self, ndarray recarr, hsize_t nrecords, int sense):
    """Converts columns in 'recarr' between NumPy and HDF5 formats.

    NumPy to HDF5 conversion is performed when 'sense' is 0.  Otherwise, HDF5
    to NumPy conversion is performed.  The conversion is done in place,
    i.e. 'recarr' is modified.

    """

    # For reading, first swap the byteorder by hand
    # (this is not currently supported by HDF5)
    if sense == 1:
      for colpathname in self.colpathnames:
        if self.coltypes[colpathname] in ["time32", "time64"]:
          colobj = self.coldescrs[colpathname]
          if hasattr(colobj, "_byteorder"):
            if colobj._byteorder != platform_byteorder:
              column = get_nested_field(recarr, colpathname)
              # Do an *inplace* byteswapping
              column.byteswap(True)

    # This should be generalised to support other type conversions.
    for t64cname in self._time64colnames:
      column = get_nested_field(recarr, t64cname)
      self._convert_time64_(column, nrecords, sense)

  def _open_append(self, ndarray recarr):
    self._v_recarray = <object>recarr
    # Get the pointer to the buffer data area
    self.wbuf = PyArray_DATA(recarr)

  def _append_records(self, hsize_t nrecords):
    cdef int ret
    cdef hsize_t nrows

    # Convert some NumPy types to HDF5 before storing.
    self._convert_types(self._v_recarray, nrecords, 0)

    nrows = self.nrows
    cdef hbool_t blosc2_support = ((self.byteorder == sys.byteorder) and
                                   (self.filters.complib != None) and
                                   (self.filters.complib[0:6] == "blosc2"))
    # release GIL (allow other threads to use the Python interpreter)
    with nogil:
        # Append the records:
        ret = H5TBOappend_records(blosc2_support, self.dataset_id,
                                  self.type_id, nrows, nrecords, self.wbuf)

    if ret < 0:
      raise HDF5ExtError("Problems appending the records.")

    self.nrows = self.nrows + nrecords

  def _close_append(self):
    cdef hsize_t nrows

    if self._v_file.params['PYTABLES_SYS_ATTRS']:
      # Update the NROWS attribute
      nrows = self.nrows
      if (H5ATTRset_attribute(self.dataset_id, "NROWS", H5T_STD_I64,
                              0, NULL, <char *>&nrows) < 0):
        raise HDF5ExtError("Problems setting the NROWS attribute.")

    # Set the caches to dirty (in fact, and for the append case,
    # it should be only the caches based on limits, but anyway)
    self._dirtycache = True
    # Delete the reference to recarray as we doesn't need it anymore
    self._v_recarray = None

  def _update_records(self, hsize_t start, hsize_t stop,
                      hsize_t step, ndarray recarr):
    cdef herr_t ret
    cdef void *rbuf
    cdef hsize_t nrecords, nrows

    # Get the pointer to the buffer data area
    rbuf = PyArray_DATA(recarr)

    # Compute the number of records to update
    nrecords = len(recarr)
    nrows = get_len_of_range(start, stop, step)
    if nrecords > nrows:
      nrecords = nrows

    # Convert some NumPy types to HDF5 before storing.
    self._convert_types(recarr, nrecords, 0)
    # Update the records:
    cdef hbool_t blosc2_support = ((self.byteorder == sys.byteorder) and
                                   (self.filters.complib != None) and
                                   (self.filters.complib[0:6] == "blosc2") and
                                   (step == 1))
    with nogil:
        ret = H5TBOwrite_records(blosc2_support, self.dataset_id,
                                 self.type_id, start, nrecords, step, rbuf )

    if ret < 0:
      raise HDF5ExtError("Problems updating the records.")

    # Set the caches to dirty
    self._dirtycache = True

  def _update_elements(self, hsize_t nrecords, ndarray coords,
                       ndarray recarr):
    cdef herr_t ret
    cdef void *rbuf
    cdef void *rcoords

    # Get the chunk of the coords that correspond to a buffer
    rcoords = PyArray_DATA(coords)

    # Get the pointer to the buffer data area
    rbuf = PyArray_DATA(recarr)

    # Convert some NumPy types to HDF5 before storing.
    self._convert_types(recarr, nrecords, 0)

    # Update the records:
    with nogil:
        ret = H5TBOwrite_elements(self.dataset_id, self.type_id,
                                  nrecords, rcoords, rbuf)

    if ret < 0:
      raise HDF5ExtError("Problems updating the records.")

    # Set the caches to dirty
    self._dirtycache = True

  def _read_records(self, hsize_t start, hsize_t nrecords, ndarray recarr):
    cdef void *rbuf
    cdef int ret
    cdef bytes fname = self._v_file.filename.encode('utf8')
    cdef char* filename = fname

    # Correct the number of records to read, if needed
    if (start + nrecords) > self.nrows:
      nrecords = self.nrows - start

    # Get the pointer to the buffer data area
    rbuf = PyArray_DATA(recarr)

    # Read the records from disk
    cdef hbool_t blosc2_support = ((self.byteorder == sys.byteorder) and
                                   (self.filters.complib != None) and
                                   (self.filters.complib[0:6] == "blosc2") and
                                   ((platform.system().lower() != 'windows') or
                                    ((self._v_file.mode == 'r'))))

    with nogil:
        ret = H5TBOread_records(filename, blosc2_support, self.dataset_id,
                                self.type_id, start, nrecords, rbuf)

    if ret < 0:
      raise HDF5ExtError("Problems reading records.")

    # Convert some HDF5 types to NumPy after reading.
    self._convert_types(recarr, nrecords, 1)

    return nrecords

  cdef hsize_t _read_chunk(self, hsize_t nchunk, ndarray iobuf, long cstart):
    cdef long nslot
    cdef hsize_t start, nrecords, chunkshape
    cdef int ret
    cdef void *rbuf
    cdef NumCache chunkcache
    cdef bytes fname = self._v_file.filename.encode('utf8')
    cdef char* filename = fname
    cdef hbool_t blosc2_support = ((self.byteorder == sys.byteorder) and
                                   (self.filters.complib != None) and
                                   (self.filters.complib[0:6] == "blosc2") and
                                   ((platform.system().lower() != 'windows') or
                                    ((self._v_file.mode == 'r'))))

    chunkcache = self._chunkcache
    chunkshape = chunkcache.slotsize
    # Correct the number of records to read, if needed
    start = nchunk*chunkshape
    nrecords = chunkshape
    if (start + nrecords) > self.nrows:
      nrecords = self.nrows - start
    rbuf = PyArray_BYTES(iobuf) + cstart * chunkcache.itemsize
    # Try to see if the chunk is in cache
    nslot = chunkcache.getslot_(nchunk)
    if nslot >= 0:
      chunkcache.getitem_(nslot, rbuf, 0)
    else:
      # Chunk is not in cache. Read it and put it in the LRU cache.
      with nogil:
          ret = H5TBOread_records(filename, blosc2_support, self.dataset_id,
                                  self.type_id, start, nrecords, rbuf)

      if ret < 0:
        raise HDF5ExtError("Problems reading chunk records.")
      nslot = chunkcache.setitem_(nchunk, rbuf, 0)
    return nrecords

  def _read_elements(self, ndarray coords, ndarray recarr):
    cdef long nrecords
    cdef void *rbuf
    cdef void *rbuf2
    cdef int ret

    # Get the chunk of the coords that correspond to a buffer
    nrecords = coords.size
    # Get the pointer to the buffer data area
    rbuf = PyArray_DATA(recarr)
    # Get the pointer to the buffer coords area
    rbuf2 = PyArray_DATA(coords)

    with nogil:
        ret = H5TBOread_elements(self.dataset_id, self.type_id,
                                 nrecords, rbuf2, rbuf)

    if ret < 0:
      raise HDF5ExtError("Problems reading records.")

    # Convert some HDF5 types to NumPy after reading.
    self._convert_types(recarr, nrecords, 1)

    return nrecords

  def _remove_rows(self, hsize_t start, hsize_t stop, long step):
    cdef size_t rowsize
    cdef hsize_t nrecords=0, nrecords2
    cdef hsize_t i
    cdef bytes fname = self._v_file.filename.encode('utf8')
    cdef char* filename = fname
    cdef hbool_t blosc2_support = ((self.byteorder == sys.byteorder) and
                                   (self.filters.complib != None) and
                                   (self.filters.complib[0:6] == "blosc2") and
                                   ((platform.system().lower() != 'windows') or
                                    ((self._v_file.mode == 'r'))))

    if step == 1:
      nrecords = stop - start
      rowsize = self.rowsize
      # Using self.disk_type_id should be faster (i.e. less conversions)
      if (H5TBOdelete_records(filename, blosc2_support, self.dataset_id,
                              self.disk_type_id, self.nrows, rowsize,
                              start, nrecords, self.nrowsinbuf) < 0):
        raise HDF5ExtError("Problems deleting records.")

      self.nrows = self.nrows - nrecords
      if self._v_file.params['PYTABLES_SYS_ATTRS']:
        # Attach the NROWS attribute
        nrecords2 = self.nrows
        H5ATTRset_attribute(self.dataset_id, "NROWS", H5T_STD_I64,
                            0, NULL, <char *>&nrecords2)
      # Set the caches to dirty
      self._dirtycache = True
    elif step == -1:
      nrecords = self._remove_rows(stop+1, start+1, 1)
    elif step >= 1:
      # always want to go through the space backwards
      for i in range(stop - step, <ssize_t>start - step, -step):
        nrecords += self._remove_rows(i, i+1, 1)
    elif step <= -1:
      # always want to go through the space backwards
      for i in range(start, stop, step):
        nrecords += self._remove_rows(i, i+1, 1)
    else:
      raise ValueError("step size may not be 0.")

    # Return the number of records removed
    return nrecords

cdef class Row:
  """Table row iterator and field accessor.

  Instances of this class are used to fetch and set the values
  of individual table fields.  It works very much like a dictionary,
  where keys are the pathnames or positions (extended slicing is
  supported) of the fields in the associated table in a specific row.

  This class provides an *iterator interface*
  so that you can use the same Row instance to
  access successive table rows one after the other.  There are also
  some important methods that are useful for accessing, adding and
  modifying values in tables.

  .. rubric:: Row attributes

  .. attribute:: nrow

      The current row number.

      This property is useful for knowing which row is being dealt with in the
      middle of a loop or iterator.

  """

  cdef npy_intp _stride
  cdef long _row, _unsaved_nrows, _mod_nrows
  cdef hsize_t start, absstep
  cdef long long stop, step, nextelement, _nrow, stopb  # has to be long long, not hsize_t, for negative step sizes
  cdef hsize_t nrowsinbuf, nrows, nrowsread
  cdef hsize_t chunksize, nchunksinbuf, totalchunks
  cdef hsize_t startb, lenbuf
  cdef long long indexchunk
  cdef int     bufcounter, counter
  cdef int     exist_enum_cols
  cdef int     _riterator, _rowsize, _write_to_seqcache
  cdef int     wherecond, indexed
  cdef int     ro_filemode, chunked
  cdef int     _bufferinfo_done, sss_on
  cdef int     iterseq_max_elements
  cdef ndarray bufcoords, indexvalid, indexvalues, chunkmap
  cdef hsize_t *bufcoords_data
  cdef hsize_t *index_values_data
  cdef char    *chunkmap_data
  cdef char    *index_valid_data
  cdef object  dtype
  cdef object  iobuf, iobufcpy
  cdef object  wrec, wreccpy
  cdef object  wfields, rfields
  cdef object  coords
  cdef object  condfunc, condargs, condkwargs
  cdef object  mod_elements, colenums
  cdef object  rfieldscache, wfieldscache
  cdef object  iterseq
  cdef object  _table_file, _table_path
  cdef object  modified_fields
  cdef object  seqcache_key

  # The nrow() method has been converted into a property, which is handier
  property nrow:
    """The current row number.

    This property is useful for knowing which row is being dealt with in the
    middle of a loop or iterator.

    """

    def __get__(self):
      return SizeType(self._nrow)

  property table:
    def __get__(self):
        self._table_file._check_open()
        return self._table_file._get_node(self._table_path)

  def __cinit__(self, table):
    cdef int nfields, i
    # Location-dependent information.
    self._table_file = table._v_file
    self._table_path = table._v_pathname
    self._unsaved_nrows = 0
    self._mod_nrows = 0
    self._row = 0
    self._nrow = 0   # Useful in mod_append read iterators
    self._riterator = 0
    self._bufferinfo_done = 0
    # Some variables from table will be cached here
    if table._v_file.mode == 'r':
      self.ro_filemode = 1
    else:
      self.ro_filemode = 0
    self.chunked = table._chunked
    self.colenums = table._colenums
    self.exist_enum_cols = len(self.colenums)
    self.nrowsinbuf = table.nrowsinbuf
    self.chunksize = table.chunkshape[0]
    self.nchunksinbuf = self.nrowsinbuf // self.chunksize
    self.dtype = table._v_dtype
    self._new_buffer(table)
    self.mod_elements = None
    self.rfieldscache = {}
    self.wfieldscache = {}
    self.modified_fields = set()

  def _iter(self, start=0, stop=0, step=1, coords=None, chunkmap=None):
    """Return an iterator for traversiong the data in table."""
    self._init_loop(start, stop, step, coords, chunkmap)
    return iter(self)

  def __iter__(self):
    """Iterator that traverses all the data in the Table"""
    return self

  cdef _new_buffer(self, table):
    """Create the recarrays for I/O buffering"""

    wdflts = table._v_wdflts
    if wdflts is None:
      self.wrec = numpy.zeros(1, dtype=self.dtype)  # Defaults are zero
    else:
      self.wrec = table._v_wdflts.copy()
    self.wreccpy = self.wrec.copy()  # A copy of the defaults
    # Build the wfields dictionary for faster access to columns
    self.wfields = {}
    for name in self.dtype.names:
      self.wfields[name] = self.wrec[name]

    # Get the read buffer for this instance (it is private, remember!)
    buff = self.iobuf = table._get_container(self.nrowsinbuf)
    # Build the rfields dictionary for faster access to columns
    # This is quite fast, as it only takes around 5 us per column
    # in my laptop (Pentium 4 @ 2 GHz).
    # F. Alted 2006-08-18
    self.rfields = {}
    for i, name in enumerate(self.dtype.names):
      self.rfields[i] = buff[name]
      self.rfields[name] = buff[name]

    # Get the stride of these buffers
    self._stride = PyArray_STRIDE(buff, 0)
    # The rowsize
    self._rowsize = self.dtype.itemsize
    self.nrows = table.nrows  # This value may change

  cdef _init_loop(self, hsize_t start, long long stop, long long step,
                 object coords, object chunkmap):
    """Initialization for the __iter__ iterator"""
    table = self.table
    self._riterator = 1   # We are inside a read iterator
    self.start = start
    self.stop = stop
    self.step = step
    self.coords = coords
    self.startb = 0
    if step > 0:
        self._row = -1  # a sentinel
        self.nrowsread = start
    elif step < 0:
        self._row = 0
        self.nrowsread = 0
        self.nextelement = start
    self._nrow = start - self.step
    self.wherecond = 0
    self.indexed = 0

    self.nrows = table.nrows   # Update the row counter

    if coords is not None and 0 < step:
      self.nrowsread = start
      self.nextelement = start
      self.stop = min(stop, len(coords))
      self.absstep = abs(step)
      return
    elif coords is not None and 0 > step:
      #self.nrowsread = 0
      #self.nextelement = start
      #self.stop = min(stop, len(coords))
      #self.stop = max(stop, start - len(coords))
      self.absstep = abs(step)
      return

    if table._where_condition:
      self.wherecond = 1
      #self.condkwargs = {'ex_uses_vml': True}
      self.condfunc, self.condargs, self.condkwargs = table._where_condition
      table._where_condition = None

    if table._use_index:
      # Indexing code depends on this condition (see #319)
      assert self.nrowsinbuf % self.chunksize == 0
      self.indexed = 1
      # Compute totalchunks here because self.nrows can change during the
      # life of a Row instance.
      self.totalchunks = self.nrows // self.chunksize
      if self.nrows % self.chunksize:
        self.totalchunks = self.totalchunks + 1
      self.nrowsread = 0
      self.nextelement = 0
      self.chunkmap = chunkmap
      self.chunkmap_data = PyArray_BYTES(self.chunkmap)
      table._use_index = False
      self.lenbuf = self.nrowsinbuf
      # Check if we have limitations on start, stop, step
      self.sss_on = (self.start > 0 or self.stop < self.nrows or self.step > 1)

    self.seqcache_key = table._seqcache_key
    table._seqcache_key = None
    if self.seqcache_key is not None:
      self._write_to_seqcache = 1
      self.iterseq_max_elements = table._v_file.params['ITERSEQ_MAX_ELEMENTS']
      self.iterseq = [] # all the row indexes, unless it would be longer than ITERSEQ_MAX_ELEMENTS
    else:
      self._write_to_seqcache = 0
      self.iterseq = None

  def __next__(self):
    """next() method for __iter__() that is called on each iteration"""

    if not self._riterator:
      # The iterator is already exhausted!
      raise StopIteration
    if self.indexed:
      return self.__next__indexed()
    elif self.coords is not None:
      return self.__next__coords()
    elif self.wherecond:
      return self.__next__inkernel()
    else:
      return self.__next__general()

  cdef __next__indexed(self):
    """The version of next() for indexed columns and a chunkmap."""

    cdef long recout, j, cs, vlen, rowsize
    cdef hsize_t nchunksread
    cdef object tmp_range
    cdef Table table
    cdef ndarray iobuf
    cdef void *IObufData
    cdef long nslot
    cdef object seq
    cdef object seqcache

    assert self.nrowsinbuf >= self.chunksize
    while self.nextelement < self.stop:
      if self.nextelement >= self.nrowsread:
        # Skip until there is interesting information
        while self.start > self.nrowsread + self.nrowsinbuf:
          self.nrowsread = self.nrowsread + self.nrowsinbuf
          self.nextelement = self.nextelement + self.nrowsinbuf

        table = self.table
        iobuf = self.iobuf
        j = 0;  recout = 0;  cs = self.chunksize
        nchunksread = self.nrowsread // cs
        tmp_range = numpy.arange(0, cs, dtype='int64')
        self.bufcoords = numpy.empty(self.nrowsinbuf, dtype='int64')
        # Fetch valid chunks until the I/O buffer is full
        while nchunksread < self.totalchunks:
          if self.chunkmap_data[nchunksread]:
            self.bufcoords[j*cs:(j+1)*cs] = tmp_range + self.nrowsread
            # Not optimized read
            #  recout = recout + table._read_records(
            #    nchunksread*cs, cs, iobuf[j*cs:])
            #
            # Optimized read through the use of a chunk cache.  This cache has
            # more or less the same speed than the integrated HDF5 chunk
            # cache, but using the PyTables one has the advantage that the
            # user can easily change this parameter.
            recout = recout + table._read_chunk(nchunksread, iobuf, j*cs)
            j = j + 1
          self.nrowsread = (nchunksread+1)*cs
          if self.nrowsread > self.stop:
            self.nrowsread = self.stop
            break
          elif j == self.nchunksinbuf:
            break
          nchunksread = nchunksread + 1

        # Evaluate the condition on this table fragment.
        iobuf = iobuf[:recout]

        if len(iobuf) > 0:
          self.table._convert_types(iobuf, len(iobuf), 1)
        self.indexvalid = call_on_recarr(
          self.condfunc, self.condargs, iobuf, **self.condkwargs)
        self.index_valid_data = PyArray_BYTES(self.indexvalid)
        # Get the valid coordinates
        self.indexvalues = self.bufcoords[:recout][self.indexvalid]
        self.index_values_data = <hsize_t *>PyArray_DATA(self.indexvalues)
        self.lenbuf = self.indexvalues.size
        # Place the valid results at the beginning of the buffer
        iobuf[:self.lenbuf] = iobuf[self.indexvalid]

        # Initialize the internal buffer row counter
        self._row = -1

        if self._write_to_seqcache:
          # Feed the indexvalues into the seqcache
          seqcache = self.iterseq
          if self.lenbuf + len(seqcache) < self.iterseq_max_elements:
            seqcache.extend(self.indexvalues)
          else:
            self.iterseq = None
            self._write_to_seqcache = 0

      self._row = self._row + 1
      # Check whether we have read all the rows in buf
      if self._row == self.lenbuf:
        self.nextelement = self.nrowsread
        # Make _row to point to the last valid entry in buffer
        # (this is useful for accessing the last row after an iterator loop)
        self._row = self._row - 1
        continue
      self._nrow = self.index_values_data[self._row]
      # Check additional conditions on start, stop, step params
      if self.sss_on:
        if (self._nrow < self.start or self._nrow >= self.stop):
          self.nextelement = self.nextelement + 1
          continue
        if (self.step > 1 and
            ((self._nrow - self.start) % self.step > 0)):
          self.nextelement = self.nextelement + 1
          continue
      # Return this row
      self.nextelement = self._nrow + 1
      return self
    else:
      # All the elements have been read for this mode
      self._finish_riterator()

  cdef __next__coords(self):
    """The version of next() for user-required coordinates"""
    cdef int recout
    cdef long long lenbuf, nextelement
    cdef object tmp
    if 0 < self.step:
      while self.nextelement < self.stop:
        if self.nextelement >= self.nrowsread:
          # Correction for avoiding reading past self.stop
          if self.nrowsread+self.nrowsinbuf > self.stop:
            lenbuf = self.stop-self.nrowsread
          else:
            lenbuf = self.nrowsinbuf
          tmp = self.coords[self.nrowsread:self.nrowsread+lenbuf:self.step]
          # We have to get a contiguous buffer, so numpy.array is the way to go
          self.bufcoords = numpy.array(tmp, dtype="uint64")
          self._row = -1
          if self.bufcoords.size > 0:
            recout = self.table._read_elements(self.bufcoords, self.iobuf)
          else:
            recout = 0
          self.bufcoords_data = <hsize_t*>PyArray_DATA(self.bufcoords)
          self.nrowsread = self.nrowsread + lenbuf
          if recout == 0:
            # no items were read, skip out
            continue
        self._row = self._row + 1
        self._nrow = self.bufcoords_data[self._row]
        self.nextelement = self.nextelement + self.absstep
        return self
      else:
        # All the elements have been read for this mode
        self._finish_riterator()
    elif 0 > self.step:
      #print("self.nextelement = ", self.nextelement, self.start, self.nrowsread, self.nextelement <  self.start - self.nrowsread + 1)
      while self.nextelement > self.stop:
        if self.nextelement < self.start - (<long long> self.nrowsread) + 1:
          if 0 > self.nextelement - (<long long> self.nrowsinbuf) + 1:
            tmp = self.coords[0:self.nextelement + 1]
          else:
            tmp = self.coords[self.nextelement - (<long long> self.nrowsinbuf) + 1:self.nextelement + 1]
          self.bufcoords = numpy.array(tmp, dtype="uint64")
          recout = self.table._read_elements(self.bufcoords, self.iobuf)
          self.bufcoords_data = <hsize_t*>PyArray_DATA(self.bufcoords)
          self.nrowsread = self.nrowsread + self.nrowsinbuf
          self._row = len(self.bufcoords) - 1
        else:
          self._row = (self._row + self.step) % len(self.bufcoords)

        self._nrow = self.nextelement - self.step
        self.nextelement = self.nextelement + self.step
        # Return this value
        return self
      else:
        # All the elements have been read for this mode
        self._finish_riterator()
    else:
      self._finish_riterator()

  cdef __next__inkernel(self):
    """The version of next() in case of in-kernel conditions"""

    cdef hsize_t recout, correct
    cdef object numexpr_locals, colvar, col
    self.nextelement = self._nrow + self.step
    while self.nextelement < self.stop:
      if self.nextelement >= self.nrowsread:
        # Skip until there is interesting information
        while self.nextelement >= self.nrowsread + self.nrowsinbuf:
          self.nrowsread = self.nrowsread + self.nrowsinbuf
        # Compute the end for this iteration
        self.stopb = self.stop - self.nrowsread
        if self.stopb > self.nrowsinbuf:
          self.stopb = self.nrowsinbuf
        self._row = self.startb - self.step
        # Read a chunk
        recout = self.table._read_records(self.nextelement, self.nrowsinbuf,
                                          self.iobuf)
        self.nrowsread = self.nrowsread + recout
        self.indexchunk = -self.step

        # Evaluate the condition on this table fragment.
        self.indexvalid = call_on_recarr(
          self.condfunc, self.condargs, self.iobuf[:recout], **self.condkwargs)
        self.index_valid_data = PyArray_BYTES(self.indexvalid)

        # Is there any interesting information in this buffer?
        if not numpy.sometrue(self.indexvalid):
          # No, so take the next one
          if self.step >= self.nrowsinbuf:
            self.nextelement = self.nextelement + self.step
          else:
            self.nextelement = self.nextelement + self.nrowsinbuf
            # Correction for step size > 1
            if self.step > 1:
              correct = (self.nextelement - self.start) % self.step
              self.nextelement = self.nextelement - correct
          continue

      self._row = self._row + self.step
      self._nrow = self.nextelement
      if self._row + self.step >= self.stopb:
        # Compute the start row for the next buffer
        self.startb = 0

      self.nextelement = self._nrow + self.step
      # Return only if this value is interesting
      self.indexchunk = self.indexchunk + self.step
      if self.index_valid_data[self.indexchunk]:
        return self
    else:
      self._finish_riterator()

  cdef __next__general(self):
    """The version of next() for the general cases"""
    cdef int recout
    if 0 < self.step:
      self.nextelement = self._nrow + self.step
      while self.nextelement < self.stop:
        if self.nextelement >= self.nrowsread:
          # Skip until there is interesting information
          while self.nextelement >= self.nrowsread + self.nrowsinbuf:
            self.nrowsread = self.nrowsread + self.nrowsinbuf
          # Compute the end for this iteration
          self.stopb = self.stop - self.nrowsread
          if self.stopb > self.nrowsinbuf:
            self.stopb = self.nrowsinbuf
          self._row = self.startb - self.step
          # Read a chunk
          recout = self.table._read_records(self.nrowsread, self.nrowsinbuf,
                                            self.iobuf)
          self.nrowsread = self.nrowsread + recout

        self._row = self._row + self.step
        self._nrow = self.nextelement
        if self._row + self.step >= self.stopb:
          # Compute the start row for the next buffer
          self.startb = (self._row + self.step) % self.nrowsinbuf

        self.nextelement = self._nrow + self.step
        # Return this value
        return self
      else:
        self._finish_riterator()
    elif 0 > self.step:
      self.stopb = -1
      while self.nextelement - 1 > self.stop:
        if self.nextelement < self.start - self.nrowsread + 1:
          # Read a chunk
          recout = self.table._read_records(self.nextelement - self.nrowsinbuf + 1,
                                            self.nrowsinbuf, self.iobuf)
          self.nrowsread = self.nrowsread + self.nrowsinbuf
          self._row = self.nrowsinbuf - 1
        else:
          self._row = (self._row + self.step) % self.nrowsinbuf

        self._nrow = self.nextelement - self.step
        self.nextelement = self.nextelement + self.step
        # Return this value
        return self
      else:
        self._finish_riterator()

  cdef _finish_riterator(self):
    """Clean-up things after iterator has been done"""
    cdef ObjectCache seqcache

    self.rfieldscache = {}     # empty rfields cache
    self.wfieldscache = {}     # empty wfields cache
    # Make a copy of the last read row in the private record
    # (this is useful for accessing the last row after an iterator loop)
    if self._row >= 0:
      self.wrec[:] = self.iobuf[self._row]
    if self._write_to_seqcache:
      seqcache = self.table._seqcache
      # Guessing iterseq size: Each element in self.iterseq should take at least 8 bytes
      seqcache.setitem_(self.seqcache_key, self.iterseq, len(self.iterseq) * 8)
    self._riterator = 0        # out of iterator
    self.iterseq = None        # empty seqcache-related things
    self.seqcache_key = None
    if self._mod_nrows > 0:    # Check if there is some modified row
      self._flush_mod_rows()     # Flush any possible modified row
    self.modified_fields = set()  # Empty the set of modified fields
    raise StopIteration        # end of iteration

  def _fill_col(self, result, start, stop, step, field):
    """Read a field from a table on disk and put the result in result"""

    cdef hsize_t startr, istartb
    cdef hsize_t istart, inrowsinbuf, inextelement
    cdef long long stopr, istopb, i, j, inrowsread
    cdef long long istop, istep
    cdef object fields

    # We can't reuse existing buffers in this context
    self._init_loop(start, stop, step, None, None)
    istart, istop, istep = self.start, self.stop, self.step
    inrowsinbuf, inextelement, inrowsread = self.nrowsinbuf, istart, istart
    istartb, startr = self.startb, 0
    i = istart
    if 0 < istep:
      while i < istop:
        if (inextelement >= inrowsread + inrowsinbuf):
          inrowsread = inrowsread + inrowsinbuf
          i = i + inrowsinbuf
          continue
        # Compute the end for this iteration
        istopb = istop - inrowsread
        if istopb > inrowsinbuf:
          istopb = inrowsinbuf
        stopr = startr + ((istopb - istartb - 1) // istep) + 1
        # Read a chunk
        inrowsread = inrowsread + self.table._read_records(i, inrowsinbuf,
                                                           self.iobuf)
        # Assign the correct part to result
        fields = self.iobuf
        if field:
          fields = get_nested_field(fields, field)
        result[startr:stopr] = fields[istartb:istopb:istep]

        # Compute some indexes for the next iteration
        startr = stopr
        j = istartb + ((istopb - istartb - 1) // istep) * istep
        istartb = (j+istep) % inrowsinbuf
        inextelement = inextelement + istep
        i = i + inrowsinbuf
    elif istep < 0:
      inrowsinbuf = self.nrowsinbuf
      #istartb = self.startb
      istartb = self.nrowsinbuf - 1
      #istopb = self.stopb - 1
      istopb = -1
      startr = 0
      i = istart
      inextelement = istart
      inrowsread = 0
      while i-1 > istop:
        #if (inextelement <= inrowsread + inrowsinbuf):
        if (inextelement < i - inrowsinbuf):
          inrowsread = inrowsread + inrowsinbuf
          i = i - inrowsinbuf
          continue
        # Compute the end for this iteration
        # (we know we are going backward so try to keep indices positive)
        stopr = startr + (1 - istopb + istartb) // (-istep)
        # Read a chunk
        inrowsread = inrowsread + self.table._read_records(i - inrowsinbuf + 1,
                                                           inrowsinbuf, self.iobuf)
        # Assign the correct part to result
        fields = self.iobuf
        if field:
          fields = get_nested_field(fields, field)
        if istopb >= 0:
            result[startr:stopr] = fields[istartb:istopb:istep]
        else:
            result[startr:stopr] = fields[istartb::istep]

        # Compute some indexes for the next iteration
        startr = stopr
        istartb = (i - istartb)%inrowsinbuf
        inextelement = inextelement + istep
        i = i - inrowsinbuf
    self._riterator = 0  # out of iterator
    return


  def append(self):
    """Add a new row of data to the end of the dataset.

    Once you have filled the proper fields for the current
    row, calling this method actually appends the new data to the
    *output buffer* (which will eventually be
    dumped to disk).  If you have not set the value of a field, the
    default value of the column will be used.

    .. warning::

        After completion of the loop in which :meth:`Row.append` has
        been called, it is always convenient to make a call to
        :meth:`Table.flush` in order to avoid losing the last rows that
        may still remain in internal buffers.

    Examples
    --------

    ::

        row = table.row
        for i in xrange(nrows):
            row['col1'] = i-1
            row['col2'] = 'a'
            row['col3'] = -1.0
            row.append()
        table.flush()

    """
    cdef ndarray iobuf, wrec, wreccpy

    if self.ro_filemode:
      raise IOError("Attempt to write over a file opened in read-only mode")

    if not self.chunked:
      raise HDF5ExtError("You cannot append rows to a non-chunked table.",
                         h5tb=False)

    if self._riterator:
      raise NotImplementedError("You cannot append rows when in middle of a table iterator. If what you want is to update records, use Row.update() instead.")

    # Commit the private record into the write buffer
    # self.iobuf[self._unsaved_nrows] = self.wrec
    # The next is faster
    iobuf = <ndarray>self.iobuf; wrec = <ndarray>self.wrec
    memcpy(PyArray_BYTES(iobuf) + self._unsaved_nrows * self._stride,
           PyArray_BYTES(wrec), self._rowsize)
    # Restore the defaults for the private record
    # self.wrec[:] = self.wreccpy
    # The next is faster
    wreccpy = <ndarray>self.wreccpy
    memcpy(PyArray_BYTES(wrec), PyArray_BYTES(wreccpy), self._rowsize)
    self._unsaved_nrows = self._unsaved_nrows + 1
    # When the buffer is full, flush it
    if self._unsaved_nrows == self.nrowsinbuf:
      self._flush_buffered_rows()

  def _flush_buffered_rows(self):
    if self._unsaved_nrows > 0:
      self.table._save_buffered_rows(self.iobuf, self._unsaved_nrows)
      # Reset the buffer unsaved counter
      self._unsaved_nrows = 0


  def _get_unsaved_nrows(self):
    return self._unsaved_nrows


  def update(self):
    """Change the data of the current row in the dataset.

    This method allows you to modify values in a table when you are in the
    middle of a table iterator like :meth:`Table.iterrows` or
    :meth:`Table.where`.

    Once you have filled the proper fields for the current row, calling
    this method actually changes data in the *output buffer* (which will
    eventually be dumped to disk).  If you have not set the value of a
    field, its original value will be used.

    .. warning::

        After completion of the loop in which :meth:`Row.update` has
        been called, it is always convenient to make a call to
        :meth:`Table.flush` in order to avoid losing changed rows that
        may still remain in internal buffers.

    Examples
    --------

    ::

        for row in table.iterrows(step=10):
            row['col1'] = row.nrow
            row['col2'] = 'b'
            row['col3'] = 0.0
            row.update()
        table.flush()

    which modifies every tenth row in table.  Or::

        for row in table.where('col1 > 3'):
            row['col1'] = row.nrow
            row['col2'] = 'b'
            row['col3'] = 0.0
            row.update()
        table.flush()

    which just updates the rows with values bigger than 3 in the first
    column.

    """

    cdef ndarray iobufcpy, iobuf

    if self.ro_filemode:
      raise IOError("Attempt to write over a file opened in read-only mode")

    if not self._riterator:
      raise NotImplementedError("You are only allowed to update rows through the Row.update() method if you are in the middle of a table iterator.")

    if self.mod_elements is None:
      # Initialize an array for keeping the modified elements
      # (just in case Row.update() would be used)
      self.mod_elements = numpy.empty(shape=self.nrowsinbuf, dtype=SizeType)
      # We need a different copy for self.iobuf here
      self.iobufcpy = self.iobuf.copy()

    # Add this row to the list of elements to be modified
    self.mod_elements[self._mod_nrows] = self._nrow
    # Copy the current buffer row in input to the output buffer
    # self.iobufcpy[self._mod_nrows] = self.iobuf[self._row]
    # The next is faster
    iobufcpy = <ndarray>self.iobufcpy; iobuf = <ndarray>self.iobuf
    memcpy(PyArray_BYTES(iobufcpy) + self._mod_nrows * self._stride,
           PyArray_BYTES(iobuf) + self._row * self._stride, self._rowsize)
    # Increase the modified buffer count by one
    self._mod_nrows = self._mod_nrows + 1
    # No point writing seqcache -- Table.flush will invalidate it
    # since we no longer know whether this row will meet _where_condition
    self._write_to_seqcache = 0
    # When the buffer is full, flush it
    if self._mod_nrows == self.nrowsinbuf:
      self._flush_mod_rows()

  def _flush_mod_rows(self):
    """Flush any possible modified row using Row.update()"""

    table = self.table
    # Save the records on disk
    table._update_elements(self._mod_nrows, self.mod_elements, self.iobufcpy)
    # Reset the counter of modified rows to 0
    self._mod_nrows = 0
    # Mark the modified fields' indexes as dirty.
    table._mark_columns_as_dirty(self.modified_fields)


  def __contains__(self, item):
    """__contains__(item)

    A true value is returned if item is found in current row, false
    otherwise.

    """

    return item in self.fetch_all_fields()

  # This method is twice as faster than __getattr__ because there is
  # not a lookup in the local dictionary
  def __getitem__(self, key):
    """__getitem__(key)

    Get the row field specified by the `key`.

    The key can be a string (the name of the field), an integer (the
    position of the field) or a slice (the range of field positions). When
    key is a slice, the returned value is a *tuple* containing the values
    of the specified fields.

    Examples
    --------

    ::

        res = [row['var3'] for row in table.where('var2 < 20')]

    which selects the var3 field for all the rows that fulfil the
    condition. Or::

        res = [row[4] for row in table if row[1] < 20]

    which selects the field in the *4th* position for all the rows that
    fulfil the condition. Or::

        res = [row[:] for row in table if row['var2'] < 20]

    which selects the all the fields (in the form of a *tuple*) for all the
    rows that fulfil the condition. Or::

        res = [row[1::2] for row in table.iterrows(2, 3000, 3)]

    which selects all the fields in even positions (in the form of a
    *tuple*) for all the rows in the slice [2:3000:3].

    """

    cdef long offset
    cdef ndarray field
    cdef object row, fields, fieldscache

    if self._riterator:
      # If in the middle of an iterator loop, the user probably wants to
      # access the read buffer
      fieldscache = self.rfieldscache; fields = self.rfields
      offset = <long>self._row
    else:
      # We are not in an iterator loop, so the user probably wants to access
      # the write buffer
      fieldscache = self.wfieldscache; fields = self.wfields
      offset = 0

    try:
      # Check whether this object is in the cache dictionary
      field = fieldscache[key]
    except (KeyError, TypeError):
      try:
        # Try to get it from fields (str or int keys)
        field = get_nested_field_cache(fields, key, fieldscache)
      except TypeError:
        # No luck yet. Still, the key can be a slice.
        # Fetch the complete row and convert it into a tuple
        if self._riterator:
          row = self.iobuf[self._row].copy().item()
        else:
          row = self.wrec[0].copy().item()
        # Try with __getitem__()
        return row[key]

    if PyArray_NDIM(field) == 1:
      # For an scalar it is not needed a copy (immutable object)
      return PyArray_GETITEM(field, PyArray_BYTES(field) + offset * self._stride)
    else:
      # Do a copy of the array, so that it can be overwritten by the user
      # without damaging the internal self.rfields buffer
      return field[offset].copy()

  # This is slightly faster (around 3%) than __setattr__
  def __setitem__(self, object key, object value):
    """__setitem__(key, value)

    Set the key row field to the specified value.

    Differently from its __getitem__() counterpart, in this case key can
    only be a string (the name of the field). The changes done via
    __setitem__() will not take effect on the data on disk until any of the
    :meth:`Row.append` or :meth:`Row.update` methods are called.

    Examples
    --------

    ::

        for row in table.iterrows(step=10):
            row['col1'] = row.nrow
            row['col2'] = 'b'
            row['col3'] = 0.0
            row.update()
        table.flush()

    which modifies every tenth row in the table.

    """

    cdef int ret
    cdef long offset
    cdef ndarray field
    cdef object fields, fieldscache

    if self.ro_filemode:
      raise IOError("attempt to write over a file opened in read-only mode")

    if self._riterator:
      # If in the middle of an iterator loop, or *after*, the user
      # probably wants to access the read buffer
      fieldscache = self.rfieldscache; fields = self.rfields
      offset = <long>self._row
    else:
      # We are not in an iterator loop, so the user probably wants to access
      # the write buffer
      fieldscache = self.wfieldscache; fields = self.wfields
      offset = 0

    # Check validity of enumerated value.
    if self.exist_enum_cols:
      if key in self.colenums:
        enum = self.colenums[key]
        for cenval in numpy.asarray(value).flat:
          enum(cenval)  # raises ``ValueError`` on invalid values

    # Get the field to be modified
    field = get_nested_field_cache(fields, key, fieldscache)
    if key not in self.modified_fields:
      self.modified_fields.add(key)

    # Finally, try to set it to the value
    try:
      # Optimization for scalar values. This can optimize the writes
      # between a 10% and 100%, depending on the number of columns modified
      if PyArray_NDIM(field) == 1:
        ret = PyArray_SETITEM(field, PyArray_BYTES(field) + offset * self._stride, value)
        if ret < 0:
          PyErr_Clear()
          raise TypeError
      ##### End of optimization for scalar values
      else:
        field[offset] = value
    except TypeError:
      raise TypeError("invalid type (%s) for column ``%s``" % (type(value),
                                                               key))

  def fetch_all_fields(self):
    """Retrieve all the fields in the current row.

    Contrarily to row[:] (see :ref:`RowSpecialMethods`), this returns row
    data as a NumPy void scalar.  For instance::

        [row.fetch_all_fields() for row in table.where('col1 < 3')]

    will select all the rows that fulfill the given condition
    as a list of NumPy records.

    """

    # We need to do a cast for recognizing negative row numbers!
    if <signed long long>self._nrow < 0:
      return ("Warning: Row iterator has not been initialized for table:\n"
              "  %s\n"
              " You will normally want to use this method in iterator "
              "contexts." % self.table)

    # Always return a copy of the row so that new data that is written
    # in self.iobuf doesn't overwrite the original returned data.
    return self.iobuf[self._row].copy()

  def __str__(self):
    """Represent the record as an string"""

    # We need to do a cast for recognizing negative row numbers!
    if <signed long long>self._nrow < 0:
      return ("Warning: Row iterator has not been initialized for table:\n"
              "  %s\n"
              " You will normally want to use this object in iterator "
              "contexts." % self.table)

    tablepathname = self.table._v_pathname
    classname = self.__class__.__name__
    return "%s.row (%s), pointing to row #%d" %  (tablepathname, classname,
                                                  self._nrow)

  def __repr__(self):
    """Represent the record as an string"""

    return str(self)

## Local Variables:
## mode: python
## py-indent-offset: 2
## tab-width: 2
## fill-column: 78
## End:



import sys
from libc.stdlib cimport free


cdef extern from "H5Zlzo.h":
    int register_lzo(char **, char **)


def register_():
    cdef char *version
    cdef char *date

    if not register_lzo(&version, &date):
        return None

    compinfo = (version, date)
    free(version)
    free(date)
    return compinfo[0].decode('ascii'), compinfo[1].decode('ascii')



"""This utility helps you migrate from PyTables 2.x APIs to 3.x APIs, which
are PEP 8 compliant.

"""
import re
import sys
import argparse
from pathlib import Path

old2newnames = dict([
    # from __init__.py
    ('hdf5Version', 'hdf5_version'),                    # data
    # from array.py
    ('parentNode', 'parentnode'),                       # kwarg
    ('getEnum', 'get_enum'),
    ('_initLoop', '_init_loop'),
    ('_fancySelection', '_fancy_selection'),
    ('_checkShape', '_check_shape'),
    ('_readSlice', '_read_slice'),
    ('_readCoords', '_read_coords'),
    ('_readSelection', '_read_selection'),
    ('_writeSlice', '_write_slice'),
    ('_writeCoords', '_write_coords'),
    ('_writeSelection', '_write_selection'),
    ('_g_copyWithStats', '_g_copy_with_stats'),
    ('_c_classId', '_c_classid'),                       # attr
    # from atom.py
    ('_checkBase', '_checkbase'),
    # from attributeset.py
    ('newSet', 'newset'),                               # kwarg
    ('copyClass', 'copyclass'),                         # kwarg
    ('_g_updateNodeLocation', '_g_update_node_location'),
    ('_g_logAdd', '_g_log_add'),
    ('_g_delAndLog', '_g_del_and_log'),
    ('_v__nodeFile', '_v__nodefile'),                   # attr (private)
    ('_v__nodePath', '_v__nodepath'),                   # attr (private)
    # from carray.py
    # ('parentNode', 'parentnode'),                       # kwarg
    # from description.py
    ('_g_setNestedNamesDescr', '_g_set_nested_names_descr'),
    ('_g_setPathNames', '_g_set_path_names'),
    ('_v_colObjects', '_v_colobjects'),                 # attr
    ('_v_nestedFormats', '_v_nested_formats'),          # attr
    ('_v_nestedNames', '_v_nested_names'),              # attr
    ('_v_nestedDescr', '_v_nested_descr'),              # attr
    ('getColsInOrder', 'get_cols_in_order'),
    ('joinPaths', 'join_paths'),
    ('metaIsDescription', 'MetaIsDescription'),
    # from earray.py
    # ('parentNode', 'parentnode'),                       # kwarg
    ('_checkShapeAppend', '_check_shape_append'),
    # from expression.py
    ('_exprvarsCache', '_exprvars_cache'),              # attr (private)
    ('_requiredExprVars', '_required_expr_vars'),
    ('setInputsRange', 'set_inputs_range'),
    ('setOutput', 'set_output'),
    ('setOutputRange', 'set_output_range'),
    # from file.py
    ('_opToCode', '_op_to_code'),                       # data (private)
    ('_codeToOp', '_code_to_op'),                       # data (private)
    ('_transVersion', '_trans_version'),                # data (private)
    ('_transGroupParent', '_trans_group_parent'),       # data (private)
    ('_transGroupName', '_trans_group_name'),           # data (private)
    ('_transGroupPath', '_trans_group_path'),           # data (private)
    ('_actionLogParent', '_action_log_parent'),         # data (private)
    ('_actionLogName', '_action_log_name'),             # data (private)
    ('_actionLogPath', '_action_log_path'),             # data (private)
    ('_transParent', '_trans_parent'),                  # data (private)
    ('_transName', '_trans_name'),                      # data (private)
    ('_transPath', '_trans_path'),                      # data (private)
    ('_shadowParent', '_shadow_parent'),                # data (private)
    ('_shadowName', '_shadow_name'),                    # data (private)
    ('_shadowPath', '_shadow_path'),                    # data (private)
    ('copyFile', 'copy_file'),
    ('openFile', 'open_file'),
    ('_getValueFromContainer', '_get_value_from_container'),
    ('__getRootGroup', '__get_root_group'),
    ('rootUEP', 'root_uep'),                            # attr
    ('_getOrCreatePath', '_get_or_create_path'),
    ('_createPath', '_create_path'),
    ('createGroup', 'create_group'),
    ('createTable', 'create_table'),
    ('createArray', 'create_array'),
    ('createCArray', 'create_carray'),
    ('createEArray', 'create_earray'),
    ('createVLArray', 'create_vlarray'),
    ('createHardLink', 'create_hard_link'),
    ('createSoftLink', 'create_soft_link'),
    ('createExternalLink', 'create_external_link'),
    ('_getNode', '_get_node'),
    ('getNode', 'get_node'),
    ('isVisibleNode', 'is_visible_node'),
    ('renameNode', 'rename_node'),
    ('moveNode', 'move_node'),
    ('copyNode', 'copy_node'),
    ('removeNode', 'remove_node'),
    ('getNodeAttr', 'get_node_attr'),
    ('setNodeAttr', 'set_node_attr'),
    ('delNodeAttr', 'del_node_attr'),
    ('copyNodeAttrs', 'copy_node_attrs'),
    ('copyChildren', 'copy_children'),
    ('listNodes', 'list_nodes'),
    ('iterNodes', 'iter_nodes'),
    ('walkNodes', 'walk_nodes'),
    ('walkGroups', 'walk_groups'),
    ('_checkOpen', '_check_open'),
    ('_isWritable', '_iswritable'),
    ('_checkWritable', '_check_writable'),
    ('_checkGroup', '_check_group'),
    ('isUndoEnabled', 'is_undo_enabled'),
    ('_checkUndoEnabled', '_check_undo_enabled'),
    ('_createTransactionGroup', '_create_transaction_group'),
    ('_createTransaction', '_create_transaction'),
    ('_createMark', '_create_mark'),
    ('enableUndo', 'enable_undo'),
    ('disableUndo', 'disable_undo'),
    ('_getMarkID', '_get_mark_id'),
    ('_getFinalAction', '_get_final_action'),
    ('getCurrentMark', 'get_current_mark'),
    ('_updateNodeLocations', '_update_node_locations'),
    # from group.py
    # ('parentNode', 'parentnode'),                       # kwarg
    # ('ptFile', 'ptfile'),                               # kwarg
    ('_getValueFromContainer', '_get_value_from_container'),
    ('_g_postInitHook', '_g_post_init_hook'),
    ('_g_getChildGroupClass', '_g_get_child_group_class'),
    ('_g_getChildLeafClass', '_g_get_child_leaf_class'),
    ('_g_addChildrenNames', '_g_add_children_names'),
    ('_g_checkHasChild', '_g_check_has_child'),
    ('_f_walkNodes', '_f_walknodes'),
    ('_g_widthWarning', '_g_width_warning'),
    ('_g_refNode', '_g_refnode'),
    ('_g_unrefNode', '_g_unrefnode'),
    ('_g_copyChildren', '_g_copy_children'),
    ('_f_getChild', '_f_get_child'),
    ('_f_listNodes', '_f_list_nodes'),
    ('_f_iterNodes', '_f_iter_nodes'),
    ('_f_walkGroups', '_f_walk_groups'),
    ('_g_closeDescendents', '_g_close_descendents'),
    ('_f_copyChildren', '_f_copy_children'),
    ('_v_maxGroupWidth', '_v_max_group_width'),         # attr
    ('_v_objectID', '_v_objectid'),                     # attr
    ('_g_loadChild', '_g_load_child'),
    ('childName', 'childname'),                         # ???
    ('_c_shadowNameRE', '_c_shadow_name_re'),           # attr (private)
    # from hdf5extension.p{yx,xd}
    ('hdf5Extension', 'hdf5extension'),
    ('_getFileId', '_get_file_id'),
    ('_flushFile', '_flush_file'),
    ('_closeFile', '_close_file'),
    ('_g_listAttr', '_g_list_attr'),
    ('_g_setAttr', '_g_setattr'),
    ('_g_getAttr', '_g_getattr'),
    ('_g_listGroup', '_g_list_group'),
    ('_g_getGChildAttr', '_g_get_gchild_attr'),
    ('_g_getLChildAttr', '_g_get_lchild_attr'),
    ('_g_flushGroup', '_g_flush_group'),
    ('_g_closeGroup', '_g_close_group'),
    ('_g_moveNode', '_g_move_node'),
    ('_convertTime64', '_convert_time64'),
    ('_createArray', '_create_array'),
    ('_createCArray', '_create_carray'),
    ('_openArray', '_open_array'),
    ('_readArray', '_read_array'),
    ('_g_readSlice', '_g_read_slice'),
    ('_g_readCoords', '_g_read_coords'),
    ('_g_readSelection', '_g_read_selection'),
    ('_g_writeSlice', '_g_write_slice'),
    ('_g_writeCoords', '_g_write_coords'),
    ('_g_writeSelection', '_g_write_selection'),
    # from idxutils.py
    ('calcChunksize', 'calc_chunksize'),
    ('infinityF', 'infinityf'),                         # data
    ('infinityMap', 'infinitymap'),                     # data
    ('infType', 'inftype'),
    ('StringNextAfter', 'string_next_after'),
    ('IntTypeNextAfter', 'int_type_next_after'),
    ('BoolTypeNextAfter', 'bool_type_next_after'),
    # from index.py
    # ('parentNode', 'parentnode'),                       # kwarg
    ('defaultAutoIndex', 'default_auto_index'),         # data
    ('defaultIndexFilters', 'default_index_filters'),   # data
    ('_tableColumnPathnameOfIndex', '_table_column_pathname_of_index'),
    ('_is_CSI', '_is_csi'),
    ('is_CSI', 'is_csi'),                               # property
    ('appendLastRow', 'append_last_row'),
    ('read_sliceLR', 'read_slice_lr'),
    ('readSorted', 'read_sorted'),
    ('readIndices', 'read_indices'),
    ('_processRange', '_process_range'),
    ('searchLastRow', 'search_last_row'),
    ('getLookupRange', 'get_lookup_range'),
    ('_g_checkName', '_g_check_name'),
    # from indexes.py
    # ('parentNode', 'parentnode'),                       # kwarg
    ('_searchBin', '_search_bin'),
    # from indexesextension
    ('indexesExtension', 'indexesextension'),
    ('initRead', 'initread'),
    ('readSlice', 'read_slice'),
    ('_readIndexSlice', '_read_index_slice'),
    ('_initSortedSlice', '_init_sorted_slice'),
    ('_g_readSortedSlice', '_g_read_sorted_slice'),
    ('_readSortedSlice', '_read_sorted_slice'),
    ('getLRUbounds', 'get_lru_bounds'),
    ('getLRUsorted', 'get_lru_sorted'),
    ('_searchBinNA_b', '_search_bin_na_b'),
    ('_searchBinNA_ub', '_search_bin_na_ub'),
    ('_searchBinNA_s', '_search_bin_na_s'),
    ('_searchBinNA_us', '_search_bin_na_us'),
    ('_searchBinNA_i', '_search_bin_na_i'),
    ('_searchBinNA_ui', '_search_bin_na_ui'),
    ('_searchBinNA_ll', '_search_bin_na_ll'),
    ('_searchBinNA_ull', '_search_bin_na_ull'),
    ('_searchBinNA_e', '_search_bin_na_e'),
    ('_searchBinNA_f', '_search_bin_na_f'),
    ('_searchBinNA_d', '_search_bin_na_d'),
    ('_searchBinNA_g', '_search_bin_na_g'),
    # from leaf.py
    # ('parentNode', 'parentnode'),                       # kwarg
    ('objectID', 'object_id'),                          # property
    ('_processRangeRead', '_process_range_read'),
    ('_pointSelection', '_point_selection'),
    ('isVisible', 'isvisible'),
    ('getAttr', 'get_attr'),
    ('setAttr', 'set_attr'),
    ('delAttr', 'del_attr'),
    # from link.py
    # ('parentNode', 'parentnode'),                       # kwarg
    ('_g_getLinkClass', '_g_get_link_class'),
    # from linkextension
    ('linkExtension', 'linkextension'),
    ('_getLinkClass', '_get_link_class'),
    ('_g_createHardLink', '_g_create_hard_link'),
    # from lrucacheextension
    ('lrucacheExtension', 'lrucacheextension'),
    # from misc/enum.py
    ('_checkAndSetPair', '_check_and_set_pair'),
    ('_getContainer', '_get_container'),
    # from misc/proxydict.py
    ('containerRef', 'containerref'),                   # attr
    # from node.py
    # ('parentNode', 'parentnode'),                       # kwarg
    ('_g_logCreate', '_g_log_create'),
    ('_g_preKillHook', '_g_pre_kill_hook'),
    ('_g_checkOpen', '_g_check_open'),
    ('_g_setLocation', '_g_set_location'),
    ('_g_updateLocation', '_g_update_location'),
    ('_g_delLocation', '_g_del_location'),
    ('_g_updateDependent', '_g_update_dependent'),
    ('_g_removeAndLog', '_g_remove_and_log'),
    ('_g_logMove', '_g_log_move'),
    ('oldPathname', 'oldpathname'),                     # ??
    ('_g_copyAsChild', '_g_copy_as_child'),
    ('_f_isVisible', '_f_isvisible'),
    ('_g_checkGroup', '_g_check_group'),
    ('_g_checkNotContains', '_g_check_not_contains'),
    ('_g_maybeRemove', '_g_maybe_remove'),
    ('_f_getAttr', '_f_getattr'),
    ('_f_setAttr', '_f_setattr'),
    ('_f_delAttr', '_f_delattr'),
    ('_v_maxTreeDepth', '_v_maxtreedepth'),             # attr
    # from nodes/filenode.py
    ('newNode', 'new_node'),
    ('openNode', 'open_node'),
    ('_lineChunkSize', '_line_chunksize'),              # attr (private)
    ('_lineSeparator', '_line_separator'),              # attr (private)
    # ('getLineSeparator', 'get_line_separator'),        # dropped
    # ('setLineSeparator', 'set_line_separator'),        # dropped
    # ('delLineSeparator', 'del_line_separator'),        # dropped
    # ('lineSeparator', 'line_separator'),                # property -- dropped
    ('_notReadableError', '_not_readable_error'),
    ('_appendZeros', '_append_zeros'),
    ('getAttrs', '_get_attrs'),
    ('setAttrs', '_set_attrs'),
    ('delAttrs', '_del_attrs'),
    ('_setAttributes', '_set_attributes'),
    ('_checkAttributes', '_check_attributes'),
    ('_checkNotClosed', '_check_not_closed'),
    ('__allowedInitKwArgs', '__allowed_init_kwargs'),   # attr (private)
    ('_byteShape', '_byte_shape'),                      # attr (private)
    ('_sizeToShape', '_size_to_shape'),                 # attr (private)
    ('_vType', '_vtype'),                               # attr (private)
    ('_vShape', '_vshape'),                             # attr (private)
    # from path.py
    ('parentPath', 'parentpath'),                       # kwarg
    ('_pythonIdRE', '_python_id_re'),                   # attr (private)
    ('_reservedIdRE', '_reserved_id_re'),               # attr (private)
    ('_hiddenNameRE', '_hidden_name_re'),               # attr (private)
    ('_hiddenPathRE', '_hidden_path_re'),               # attr (private)
    ('checkNameValidity', 'check_name_validity'),
    ('joinPath', 'join_path'),
    ('splitPath', 'split_path'),
    ('isVisibleName', 'isvisiblename'),
    ('isVisiblePath', 'isvisiblepath'),
    # from registry.py
    ('className', 'classname'),                         # kwarg
    ('classNameDict', 'class_name_dict'),               # data
    ('classIdDict', 'class_id_dict'),                   # data
    ('getClassByName', 'get_class_by_name'),
    # from scripts/ptdump.py
    ('dumpLeaf', 'dump_leaf'),
    ('dumpGroup', 'dump_group'),
    # from scripts/ptrepack.py
    ('newdstGroup', 'newdst_group'),
    ('recreateIndexes', 'recreate_indexes'),
    ('copyLeaf', 'copy_leaf'),
    # from table.py
    # ('parentNode', 'parentnode'),                       # kwarg
    ('_nxTypeFromNPType', '_nxtype_from_nptype'),       # data (private)
    ('_npSizeType', '_npsizetype'),                     # data (private)
    ('_indexNameOf', '_index_name_of'),
    ('_indexPathnameOf', '_index_pathname_of'),
    ('_indexPathnameOfColumn', '_index_pathname_of_column'),
    ('_indexNameOf_', '_index_name_of_'),
    ('_indexPathnameOf_', '_index_pathname_of_'),
    ('_indexPathnameOfColumn_', '_index_pathname_of_column_'),
    ('_table__setautoIndex', '_table__setautoindex'),
    ('_table__getautoIndex', '_table__getautoindex'),
    ('_table__autoIndex', '_table__autoindex'),         # data (private)
    ('_table__whereIndexed', '_table__where_indexed'),
    ('createIndexesTable', 'create_indexes_table'),
    ('createIndexesDescr', 'create_indexes_descr'),
    ('_column__createIndex', '_column__create_index'),
    ('_autoIndex', '_autoindex'),                       # attr
    ('autoIndex', 'autoindex'),                         # attr
    ('_useIndex', '_use_index'),
    ('_whereCondition', '_where_condition'),            # attr (private)
    ('_conditionCache', '_condition_cache'),            # attr (private)
    # ('_exprvarsCache', '_exprvars_cache'),
    ('_enabledIndexingInQueries',
     '_enabled_indexing_in_queries'),  # attr (private)
    ('_emptyArrayCache', '_empty_array_cache'),         # attr (private)
    ('_getTypeColNames', '_get_type_col_names'),
    ('_getEnumMap', '_get_enum_map'),
    ('_cacheDescriptionData', '_cache_description_data'),
    ('_getColumnInstance', '_get_column_instance'),
    ('_checkColumn', '_check_column'),
    ('_disableIndexingInQueries', '_disable_indexing_in_queries'),
    ('_enableIndexingInQueries', '_enable_indexing_in_queries'),
    # ('_requiredExprVars', '_required_expr_vars'),
    ('_getConditionKey', '_get_condition_key'),
    ('_compileCondition', '_compile_condition'),
    ('willQueryUseIndexing', 'will_query_use_indexing'),
    ('readWhere', 'read_where'),
    ('whereAppend', 'append_where'),
    ('getWhereList', 'get_where_list'),
    ('_check_sortby_CSI', '_check_sortby_csi'),
    ('_readCoordinates', '_read_coordinates'),
    ('readCoordinates', 'read_coordinates'),
    ('_saveBufferedRows', '_save_buffered_rows'),
    ('modifyCoordinates', 'modify_coordinates'),
    ('modifyRows', 'modify_rows'),
    ('modifyColumn', 'modify_column'),
    ('modifyColumns', 'modify_columns'),
    ('flushRowsToIndex', 'flush_rows_to_index'),
    ('_addRowsToIndex', '_add_rows_to_index'),
    ('removeRows', 'remove_rows'),
    ('_setColumnIndexing', '_set_column_indexing'),
    ('_markColumnsAsDirty', '_mark_columns_as_dirty'),
    ('_reIndex', '_reindex'),
    ('_doReIndex', '_do_reindex'),
    ('reIndex', 'reindex'),
    ('reIndexDirty', 'reindex_dirty'),
    ('_g_copyRows', '_g_copy_rows'),
    ('_g_copyRows_optim', '_g_copy_rows_optim'),
    ('_g_propIndexes', '_g_prop_indexes'),
    ('_g_updateTableLocation', '_g_update_table_location'),
    ('_tableFile', '_table_file'),                      # attr (private)
    ('_tablePath', '_table_path'),                      # attr (private)
    ('createIndex', 'create_index'),
    ('createCSIndex', 'create_csindex'),
    ('removeIndex', 'remove_index'),
    # from tableextension
    ('tableExtension', 'tableextension'),
    ('getNestedFieldCache', 'get_nested_field_cache'),
    ('getNestedType', 'get_nested_type'),
    ('_createTable', '_create_table'),
    ('_getInfo', '_get_info'),
    ('indexChunk', 'indexchunk'),                       # attr
    ('indexValid', 'indexvalid'),                       # attr
    ('indexValues', 'indexvalues'),                     # attr
    ('bufcoordsData', 'bufcoords_data'),                # attr
    ('indexValuesData', 'index_values_data'),           # attr
    ('chunkmapData', 'chunkmap_data'),                  # attr
    ('indexValidData', 'index_valid_data'),             # attr
    ('whereCond', 'wherecond'),                         # attr
    ('iterseqMaxElements', 'iterseq_max_elements'),     # attr
    ('IObuf', 'iobuf'),                                 # attr
    ('IObufcpy', 'iobufcpy'),                           # attr
    ('_convertTime64_', '_convert_time64_'),
    ('_convertTypes', '_convert_types'),
    ('_newBuffer', '_new_buffer'),
    ('__next__inKernel', '__next__inkernel'),
    ('_fillCol', '_fill_col'),
    ('_flushBufferedRows', '_flush_buffered_rows'),
    ('_getUnsavedNrows', '_get_unsaved_nrows'),
    ('_flushModRows', '_flush_mod_rows'),
    # from undoredo.py
    ('moveToShadow', 'move_to_shadow'),
    ('moveFromShadow', 'move_from_shadow'),
    ('undoCreate', 'undo_create'),
    ('redoCreate', 'redo_create'),
    ('undoRemove', 'undo_remove'),
    ('redoRemove', 'redo_remove'),
    ('undoMove', 'undo_move'),
    ('redoMove', 'redo_move'),
    ('attrToShadow', 'attr_to_shadow'),
    ('attrFromShadow', 'attr_from_shadow'),
    ('undoAddAttr', 'undo_add_attr'),
    ('redoAddAttr', 'redo_add_attr'),
    ('undoDelAttr', 'undo_del_attr'),
    ('redoDelAttr', 'redo_del_attr'),
    # from utils.py
    ('convertToNPAtom', 'convert_to_np_atom'),
    ('convertToNPAtom2', 'convert_to_np_atom2'),
    ('checkFileAccess', 'check_file_access'),
    ('logInstanceCreation', 'log_instance_creation'),
    ('fetchLoggedInstances', 'fetch_logged_instances'),
    ('countLoggedInstances', 'count_logged_instances'),
    ('listLoggedInstances', 'list_logged_instances'),
    ('dumpLoggedInstances', 'dump_logged_instances'),
    ('detectNumberOfCores', 'detect_number_of_cores'),
    # from utilsextension
    ('utilsExtension', 'utilsextension'),
    ('PTTypeToHDF5', 'pttype_to_hdf5'),                 # data
    ('PTSpecialKinds', 'pt_special_kinds'),             # data
    ('NPExtPrefixesToPTKinds', 'npext_prefixes_to_ptkinds'),    # data
    ('HDF5ClassToString', 'hdf5_class_to_string'),      # data
    ('setBloscMaxThreads', 'set_blosc_max_threads'),
    ('silenceHDF5Messages', 'silence_hdf5_messages'),
    ('isHDF5File', 'is_hdf5_file'),
    ('isPyTablesFile', 'is_pytables_file'),
    ('getHDF5Version', 'get_hdf5_version'),
    ('getPyTablesVersion', 'get_pytables_version'),
    ('whichLibVersion', 'which_lib_version'),
    ('whichClass', 'which_class'),
    ('getNestedField', 'get_nested_field'),
    ('getFilters', 'get_filters'),
    ('getTypeEnum', 'get_type_enum'),
    ('enumFromHDF5', 'enum_from_hdf5'),
    ('enumToHDF5', 'enum_to_hdf5'),
    ('AtomToHDF5Type', 'atom_to_hdf5_type'),
    ('loadEnum', 'load_enum'),
    ('HDF5ToNPNestedType', 'hdf5_to_np_nested_type'),
    ('HDF5ToNPExtType', 'hdf5_to_np_ext_type'),
    ('AtomFromHDF5Type', 'atom_from_hdf5_type'),
    ('createNestedType', 'create_nested_type'),
    # from unimlemented.py
    ('_openUnImplemented', '_open_unimplemented'),
    # from vlarray.py
    # ('parentNode', 'parentnode'),                       # kwarg
    # ('expectedsizeinMB', 'expected_mb'),                # --> expectedrows
    # ('_v_expectedsizeinMB', '_v_expected_mb'),          # --> expectedrows
])

new2oldnames = {v: k for k, v in old2newnames.items()}

# Note that it is tempting to use the ast module here, but then this
# breaks transforming cython files.  So instead we are going to do the
# dumb thing with replace.


def make_subs(ns):
    names = new2oldnames if ns.reverse else old2newnames
    s = r'(?<=\W)({})(?=\W)'.format('|'.join(list(names)))
    if ns.ignore_previous:
        s += r'(?!\s*?=\s*?previous_api(_property)?\()'
        s += r'(?!\* to \*\w+\*)'
        s += r'(?!\* parameter has been renamed into \*\w+\*\.)'
        s += r'(?! is pending deprecation, import \w+ instead\.)'
    subs = re.compile(s, flags=re.MULTILINE)

    def repl(m):
        return names.get(m.group(1), m.group(0))
    return subs, repl


def main():
    desc = ('PyTables 2.x -> 3.x API transition tool\n\n'
            'This tool displays to standard out, so it is \n'
            'common to pipe this to another file:\n\n'
            '$ pt2to3 oldfile.py > newfile.py')
    parser = argparse.ArgumentParser(description=desc)
    parser.add_argument('-r', '--reverse', action='store_true', default=False,
                        dest='reverse',
                        help="reverts changes, going from 3.x -> 2.x.")
    parser.add_argument('-p', '--no-ignore-previous', action='store_false',
                        default=True, dest='ignore_previous',
                        help="ignores previous_api() calls.")
    parser.add_argument('-o', default=None, dest='output',
                        help="output file to write to.")
    parser.add_argument('-i', '--inplace', action='store_true', default=False,
                        dest='inplace', help="overwrites the file in-place.")
    parser.add_argument('filename', help='path to input file.')
    ns = parser.parse_args()

    if not Path(ns.filename).is_file():
        sys.exit(f'file {ns.filename!r} not found')
    src = Path(ns.filename).read_text()

    subs, repl = make_subs(ns)
    targ = subs.sub(repl, src)

    ns.output = ns.filename if ns.inplace else ns.output
    if ns.output is None:
        sys.stdout.write(targ)
    else:
        Path(ns.output).write_text(targ)


if __name__ == '__main__':
    main()



"""Special node behaviours for PyTables.

This package contains several modules that give specific behaviours
to PyTables nodes.  For instance, the filenode module provides
a file interface to a PyTables node.


Package modules:
  filenode -- A file interface to nodes for PyTables databases.

"""

# The list of names to be exported to the importing module.
__all__ = ['filenode']
